/**
 * Vector file reader.
 *
 * Loosely based on the work of Nathan Cassee (TU/e).
 *
 * Part of Master Project 'Porting tree-based hash table compression to GPGPU model checking', 
 * University of Twente (UT).
 *
 * @author  Danny Bergsma (s0113360)
 * @version 1.0
 */


	// -- INCLUDES --


#include "vector_file_reader.hpp"

#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <string>
#include <vector>


using namespace std;


	// -- LITERALS/CONSTANTS --


/** File mode used for reading the vector file. */
static const char FILE_MODE[] = "r";
/** Character that separates elements in a vector file line. */
static const char SEPARATOR   = ',';

/** Error message for no valid file. */
static const char ERROR_NO_VALID_FILE[]            = "No valid file provided\n";
/** Error message for file with vectors with different number of elements. */
static const char ERROR_VECTORS_NOT_EQUAL_LENGTH[] = "All vectors should be of equal length\n";


	// -- FUNCTIONS --


Vectors readVectorFile(const char *fileName) {
	auto result = Vectors();
	
	if (fileName != NULL) {
		FILE *file = fopen(fileName, FILE_MODE);
		if (file == NULL)
			fprintf(stderr, ERROR_NO_VALID_FILE);
		else {
			vector<vector<element_t>> vectorSequence;  // contains all vectors read so far
			
			// reading complete file, adding vectors to vectorSequence:
			vector<element_t> bufferVector;
			string            bufferElement;
			char read = fgetc(file);
			while (!feof(file)) {
				if ((read == '\n') || (read == SEPARATOR)) {  // current element completed
					bufferVector.push_back(stoul(bufferElement));
					bufferElement = "";
					if (read == '\n') {  // current vector completed
						vectorSequence.push_back(bufferVector);
						bufferVector = vector<element_t>();
					}
				} else
					bufferElement += read;
				read = fgetc(file);
			}
			if (bufferElement != "") {  // bufferElement left
				bufferVector.push_back(stoul(bufferElement));
				vectorSequence.push_back(bufferVector);
			}
			fclose(file);
			
			// converting vectorSequence to return type:
			if (vectorSequence.size() > 0) {
				result.nVectors     =                vectorSequence.size();
				result.vectorLength = (unsigned int) vectorSequence[0].size();
				result.elements     = (element_t*)   malloc((result.nVectors * result.vectorLength) * sizeof(element_t));
				
				// copying all elements from vectorSequence to result.elements (if all vectors have equal size):
				size_t vectorIndex = 0;
				while (vectorIndex < result.nVectors)
					if (vectorSequence[vectorIndex].size() == result.vectorLength) {
						memcpy((result.elements + (vectorIndex * result.vectorLength)),
						       vectorSequence[vectorIndex].data(),
						       (result.vectorLength * sizeof(*result.elements)));
						vectorIndex++;
					} else {  // vector has different number of elements
						fprintf(stderr, ERROR_VECTORS_NOT_EQUAL_LENGTH);
						result.nVectors     = 0;  // + exit loop
						result.vectorLength = 0;
						free(result.elements);
						result.elements = NULL;
					}
			}
		}
	} else
		fprintf(stderr, ERROR_NO_VALID_FILE);
	
	return result;
}