/**
 * GPU hash table implementation.
 * Compressed version (64-bits), with linear probing (32 entries).
 *
 * Loosely based on the work of Nathan Cassee (TU/e).
 *
 * Part of Master Project 'Porting tree-based hash table compression to GPGPU model checking', 
 * University of Twente (UT).
 *
 * @author  Danny Bergsma (s0113360)
 * @version 3.0
 */


	// -- INCLUDES --


#include "hash_table.h"

#include <stdio.h>

#include <cuda_runtime.h>

#include "mt64.h"


	// --- DEVICE CODE ---
	
	
		// -- INCLUDES --
	
	
	#include <device_launch_parameters.h>
	
	
		// -- ENUMS --
	
	
	/** Enumeration of possible return values of findOrPut(). */
	enum FoundOrPut {NEW, SEEN, FULL};
	
	
		// -- LITERALS/CONSTANTS --
	
		
	/** 64-bit version of the empty entry. */
	static const uint64_t EMPTY_ENTRY_64    = (((uint64_t) EMPTY_ENTRY_32 << (sizeof(EMPTY_ENTRY_32) * 8)) | EMPTY_ENTRY_32);
	/** Bitmask to unroot 64-bit entries (two 32-bit entries). */
	static const uint64_t UNROOT_BITMASK_64 = ~((uint64_t) ROOT_BIT_32);  // little-endian
	
	/** Number of hash functions (bounded double hashing). */
	static const unsigned int N_HASH_FUNCTIONS              = 32;
	/** Number of constants per hash function. */
	static const unsigned int N_CONSTANTS_PER_HASH_FUNCTION = 3;
	/** Total number of hash constants. */
	static const unsigned int N_TOTAL_HASH_CONSTANTS        = (N_HASH_FUNCTIONS * N_CONSTANTS_PER_HASH_FUNCTION);
	
	/** Size of hash in bits. */
	static const unsigned int HASH_BITS = (sizeof(uint32_t) * 8);

	/** Number of 64-bit entries in a linear probe. */
	static const unsigned int LINEAR_PROBE_LENGTH     = (WARP_SIZE / 8);
	/** Number of 64-bit entries in an external linear probe. */
	static const unsigned int LINEAR_EXT_PROBE_LENGTH = (WARP_SIZE / 2);
	
	/** Status message when table is (considered) full. */
	#define STATUS_TABLE_FULL  "Table is (considered) full\n"
	
	
		// -- CONSTANTS --
	
	
	/** Constants used in hash functions. */
	__constant__ uint64_t     d_hashConstants[N_TOTAL_HASH_CONSTANTS];
	/** Hash table: array of hash table entries. */
	__constant__ inttype      *d_table;
	/** 
	 * Number of hash table entries.
	 * @invariant length of d_table array == d_tableLength
	 * @invariant d_tableLength == (d_nBuckets * (MAX_CHILDREN_BINARY_TREE * d_bucketSize))
	 * @invariant (0 < d_tableLength <= MAX_ENTRIES) && ((d_tableLength % WARP_SIZE) == 0)
	 */
	__constant__ indextype    d_tableLength;
	/** Number of buckets. */
	__constant__ uint32_t     d_nBuckets;
	/**
	 * Number of 64-bit hash table entries that form a bucket.
	 * @invariant (0 < d_bucketSize <= WARP_SIZE) && (d_bucketSize is power of 2)
	 */
	__constant__ unsigned int d_bucketSize;
	/** Length of vector (number of elements in vectors). */
	__constant__ unsigned int d_vectorLength;
	/**
	 * Number of iterations in linear probing.
	 * @invariant d_linearProbeIterations == (LINEAR_PROBE_LENGTH / d_bucketSize)
	 */
	__constant__ unsigned int d_linearProbeIterations;
	/**
	 * Number of iterations in external linear probing.
	 * @invariant d_linearExtProbeIterations == (LINEAR_EXT_PROBE_LENGTH / d_bucketSizeMinLPB)
	 */
	__constant__ unsigned int d_linearExtProbeIterations;
	/**
	 * Number of hash table entries that form a bucket, with a minimum of LINEAR_PROBE_LENGTH.
	 * @invariant d_bucketSizeMinLPB == max(d_bucketSize, LINEAR_PROBE_LENGTH)
	 */
	__constant__ unsigned int d_bucketSizeMinLPB;
	
	
		// -- MACROS --
	
	
	/** Counts all hash table entries for which condition holds. */
	#define COUNT_DEVICE(condition)  {                                                                                      \
		indextype localCount = 0;                                                                                           \
		for (indextype entryIndex = ((((indextype) blockIdx.x) * blockDim.x) + threadIdx.x); (entryIndex < d_tableLength);  \
			  (entryIndex += (blockDim.x * gridDim.x)))                                                                     \
			if (condition)                                                                                                  \
				localCount++;                                                                                               \
		                                                                                                                    \
		atomicAdd(count, localCount);                                                                                       \
	}
	
	/** 
	 * Index of thread in warp.
	 * @invariant 0 <= LANE < WARP_SIZE
	 */
	#define LANE  (threadIdx.x % WARP_SIZE)
	
	/** 
	 * Index of bucket group (in warp) the current thread belongs to.
	 * @invariant 0 <= BUCKET_GROUP_INDEX < WARP_SIZE
	 */
	#define BUCKET_GROUP_INDEX(bucketSize)  (LANE / bucketSize)
	
	/**
	 * Index (offset) of entry (in hash table bucket) that corresponds to the current thread.
	 * @invariant 0 <= ENTRY_INDEX_IN_BUCKET < bucketSize
	 */
	#define ENTRY_INDEX_IN_BUCKET(bucketSize)  (LANE % bucketSize)
		
	/** Sets bits lower..(lower+amount-1) of bitmask to 1, all other bits to 0 (index LSB = 0). */
	#define SET_BITS(bitmask, lower, amount)  bitmask = ((uint32_t) (((1ull << (lower + amount)) - 1) ^ ((1ull << lower) - 1)));
	
	/**
	 * Returns lane+1 of the first thread in the bucket group for which the predicate holds.
	 * Returns 0 if there are no such threads.
	 * @ensures 0 <= GET_FIRST_MATCHING_THREAD <= WARP_SIZE
	 */
	#define GET_FIRST_MATCHING_THREAD(bucketGroupMask, predicate)  \
		__ffs(__ballot_sync(bucketGroupMask, predicate))
	
	
		// -- KERNEL: INITIALIZING TABLE --
	
	
	/**
	 * Initializes the hash table by setting all entries to the empty entry value, in parallel.
	 */
	__global__ void initTable() {
		for (indextype entryIndex = ((((indextype) blockIdx.x) * blockDim.x) + threadIdx.x); (entryIndex < d_tableLength);
			  (entryIndex += (blockDim.x * gridDim.x)))  // grid-stride loop
			d_table[entryIndex] = EMPTY_ENTRY_32;
	}
	
	
		// -- KERNEL: COUNTING ELEMENTS --
	
	
	/** 
	 * Counts all elements in hash table, in parallel.
	 * @param count will (eventually) be set to total number of elements in hash table
	 */
	__global__ void countElements(indextype *count) {
		COUNT_DEVICE(d_table[entryIndex] != EMPTY_ENTRY_32)
	}
	
	
		// -- KERNEL: COUNTING VECTORS --
	
	
	/**
	 * Counts all vectors in hash table, in parallel.
	 * @param count will (eventually) be set to total number of vectors in hash table
	 */
	__global__ void countVectors(indextype *count) {
		COUNT_DEVICE(d_table[entryIndex] & ROOT_BIT_32)
	}
	
	
		// -- HELPER FUNCTIONS KERNEL INSERTING VECTORS: FIND_OR_PUT --
	
	
	/**
	 * Returns hash of vector, using the hash function with index hashFunctionIndex.
	 * Hash is valid as index of hash table.
	 * @param    hashFunctionIndex index of hash function to be used
	 * @param    vector            vector to be hashed
	 * @return   hash of vector
	 * @requires hashFunctionIndex < N_HASH_FUNCTIONS
	 * @requires length of vector == MAX_CHILDREN_BINARY_TREE
	 * @ensures  0 <= \result <= (d_tableLength / MAX_CHILDREN_BINARY_TREE - d_bucketSize)
	 */
	__device__ indextype hash(unsigned int hashFunctionIndex, const inttype *vector) {
		unsigned int hashConstantsOffset = (hashFunctionIndex * N_CONSTANTS_PER_HASH_FUNCTION);

		return (((((((d_hashConstants[hashConstantsOffset] + vector[1]) * (d_hashConstants[hashConstantsOffset + 1] + vector[0]))
			+ d_hashConstants[hashConstantsOffset + 2]) >> HASH_BITS) * d_nBuckets) >> HASH_BITS) * d_bucketSize);
	}
	
	/**
	 * Tries to add vector to hash table, using (partial) warp-the-line, bounded double hashing and linear probing.
	 * Returns index of vector in hash table; EMPTY_ENTRY_32 is returned if table is full.
	 * @param    vector          (sub)vector to be inserted into table
	 * @param    bucketGroupMask bitmask of bucket group the current thread belongs to
	 * @return   index of vector in hash table; EMPTY_ENTRY_32 is returned if table is full
	 * @requires length of vector == MAX_CHILDREN_BINARY_TREE
	 * @requires (\forall int elementIndex; 0 <= elementIndex < MAX_CHILDREN_BINARY_TREE;
	 *                      (vector[elementIndex] != EMPTY_ENTRY_32))
	 *                         && !(vector[elementIndex] & ROOT_BIT_32)))
	 */
	__device__ indextype findOrPut(const inttype *vector, uint32_t bucketGroupMask) {
		indextype result = EMPTY_ENTRY_32;  // assume full table (or considered full)
		
		uint64_t vector64 = *((uint64_t*) vector);
		
		unsigned int hashFunctionIndex = 0;
		while ((result == EMPTY_ENTRY_32) && (hashFunctionIndex < N_HASH_FUNCTIONS)) {
			indextype entryIndex = (hash(hashFunctionIndex, vector) + ENTRY_INDEX_IN_BUCKET(d_bucketSize));
			
			unsigned int extProbeIndex = 0;
			do {
				unsigned int probeIndex = 0;
				do {
					uint64_t *entryAddress = (uint64_t*) (d_table + (MAX_CHILDREN_BINARY_TREE * entryIndex)),
							 entry         =             (*entryAddress & UNROOT_BITMASK_64);  // unset root bit
					
					unsigned int firstMatchingSlot = GET_FIRST_MATCHING_THREAD(bucketGroupMask, (entry == vector64));  // 0..32
					if (firstMatchingSlot)  // 1..32; vector group found vector in (slots of bucket of) hash table
						result = __shfl_sync(bucketGroupMask, entryIndex, (firstMatchingSlot - 1));  // + exit all three loops
					else {  // vector was not found ==> try to insert vector into first empty slot in bucket
						// setting firstEmptySlot to lane+1 of first vector group with empty slot:
						firstMatchingSlot = GET_FIRST_MATCHING_THREAD(bucketGroupMask, (entry == EMPTY_ENTRY_64));  // 0..32
						
						if (firstMatchingSlot--)  // firstEmptySlot is in 1..32 <-> LANE is in 0..31
							do {
								if (LANE == firstMatchingSlot)
									entry = (atomicCAS((unsigned long long*) entryAddress, EMPTY_ENTRY_64, vector64)
											 & UNROOT_BITMASK_64);  // unset root bit
								
								entry = __shfl_sync(bucketGroupMask, entry, firstMatchingSlot);
								if ((entry == EMPTY_ENTRY_64) || (entry == vector64))
								  // vector group added/found vector in (slot of) hash table
									result = __shfl_sync(bucketGroupMask, entryIndex, firstMatchingSlot);  // + exit all four loops
								else
									firstMatchingSlot++;  // try next (empty) slot in bucket
							} while ((result == EMPTY_ENTRY_32) && (firstMatchingSlot % d_bucketSize));
							  // continue till vector is inserted or there are no empty (complete) slots anymore
					}
					
					entryIndex = (((entryIndex / LINEAR_PROBE_LENGTH) * LINEAR_PROBE_LENGTH)
								  + ((entryIndex + d_bucketSize) % LINEAR_PROBE_LENGTH));
					probeIndex++;
				} while ((result == EMPTY_ENTRY_32) && (probeIndex < d_linearProbeIterations));
				
				entryIndex = (((entryIndex / LINEAR_EXT_PROBE_LENGTH) * LINEAR_EXT_PROBE_LENGTH)
				              + ((entryIndex + d_bucketSizeMinLPB) % LINEAR_EXT_PROBE_LENGTH));
				extProbeIndex++;
			} while ((result == EMPTY_ENTRY_32) && (extProbeIndex < d_linearExtProbeIterations));

			hashFunctionIndex++;  // try next hash function (double hashing)
		}
		
		return result;
	}
	
	/**
	 * Tries to add (sub)vector to hash table, using tree-based hash table compression. Returns index of vector in hash table.
	 * If table is (considered) full, EMPTY_ENTRY_32 is returned.
	 * @param    vector          (sub)vector to be inserted into table
	 * @param    vectorLength    length of (sub)vector
	 * @param    bucketGroupMask bitmask of bucket group the current thread belongs to
	 * @return   index of vector in hash table
	 * @requires vectorLength > 0
	 * @requires length of vector == vectorLength
	 * @requires (\forall int elementIndex; 0 <= elementIndex < vectorLength;
	 *                      (vector[elementIndex] != EMPTY_ENTRY_32))
	 *                         && !(vector[elementIndex] & ROOT_BIT_32)))
	 */
	__device__ indextype treeRec(const inttype *vector, unsigned int vectorLength, uint32_t bucketGroupMask) {
		indextype result = EMPTY_ENTRY_32;  // assume full table
		
		indextype recResult[MAX_CHILDREN_BINARY_TREE];
		if (vectorLength < (MAX_CHILDREN_BINARY_TREE * MAX_CHILDREN_BINARY_TREE)) {
		  // vectorLength < 4 ==> we need only one or no recursive call at all
			recResult[0] = vector[0];
			recResult[1] = ((vectorLength == 1) ? 0 
			                                    : ((vectorLength == MAX_CHILDREN_BINARY_TREE) ? vector[1]
			                                                                                  : treeRec((vector + 1), MAX_CHILDREN_BINARY_TREE, 
			                                                                                              bucketGroupMask)));
		} else {  // vectorLength >= 4 ==> two recursive calls
			unsigned int split = (vectorLength / MAX_CHILDREN_BINARY_TREE);
			recResult[0]       = treeRec( vector,          split,                  bucketGroupMask);
			recResult[1]       = treeRec((vector + split), (vectorLength - split), bucketGroupMask);
		}
		
		if ((recResult[0] != EMPTY_ENTRY_32) && (recResult[1] != EMPTY_ENTRY_32))  // check for full table
			result = findOrPut(recResult, bucketGroupMask);
		
		return result;
	}
	
	/** 
	 * Tries to add vector to hash table, using tree-based hash table compression. 
	 * If it was already present, SEEN is returned. If it was not present, the vector is added and NEW is returned.
	 * If table is (considered) full, FULL is returned.
	 * @param    vector          vector to be inserted into table
	 * @param    bucketGroupMask bitmask of bucket group the current thread belongs to
	 * @return   result: SEEN (already present), NEW (was not present) or FULL (table is full)
	 * @requires d_vectorLength > 0
	 * @requires length of vector == d_vectorLength 
	 * @requires (\forall int elementIndex; 0 <= elementIndex < d_vectorLength;
	 *                      (vector[elementIndex] != EMPTY_ENTRY_32))
	 *                         && !(vector[elementIndex] & ROOT_BIT_32)))
	 */
	__device__ FoundOrPut treeFindOrPut(const inttype *vector, uint32_t bucketGroupMask) {
		FoundOrPut result = FULL;  // assume full table
		
		indextype rootIndex = treeRec(vector, d_vectorLength, bucketGroupMask);
		if ((ENTRY_INDEX_IN_BUCKET(d_bucketSize) == 0) && (rootIndex != EMPTY_ENTRY_32))
		  // only first thread in bucket group should check and set ROOT_BIT
			result = ((atomicOr((d_table + (MAX_CHILDREN_BINARY_TREE * rootIndex)), ROOT_BIT_32) & ROOT_BIT_32) ? SEEN : NEW);
		result = (FoundOrPut) __shfl_sync(bucketGroupMask, result, 0, d_bucketSize);
		
		return result;
	}
	
	
		// -- KERNEL: INSERTING VECTORS --
	
	
	/**
	 * Inserts the given (elements of) vectors into the (initialized) hash table, in parallel.
	 * Some vectors may not be inserted, if the hash table is (considered) full.
	 * In that case a message will be printed.
	 * @param    d_elements     array of (the elements of) the vectors to be inserted
	 * @param    elementsLength total number of elements to be inserted
	 * @requires (elementsLength % vector length) == 0
	 * @requires elementsLength <= length of d_elements array
	 * @requires (\forall int elementIndex; 0 <= elementIndex < elementsLength; 
	 *                      (d_elements[elementIndex] != EMPTY_ENTRY_32))
	 *                         && !(d_elements[elementIndex] & ROOT_BIT_32)))
	 */
	__global__ void insertTable(const inttype *d_elements, size_t elementsLength) {
		uint32_t bucketGroupMask;
		/* setting bitmask to bucket group of current thread: only the bits of those threads are set
		   that are part of the bucket group of the current thread */
		SET_BITS(bucketGroupMask, (BUCKET_GROUP_INDEX(d_bucketSize) * d_bucketSize), d_bucketSize)
		
		// processing (elements of) vectors (by grid-stride loop):
		size_t vectorIndex  = (((((size_t) blockIdx.x) * blockDim.x) + threadIdx.x)  / d_bucketSize),
		       stride       = (((((size_t) gridDim.x)  * blockDim.x) / d_bucketSize) * d_vectorLength),
		       elementIndex = (vectorIndex * d_vectorLength);
		while (elementIndex < elementsLength) {
			if ((treeFindOrPut((d_elements + elementIndex), bucketGroupMask) == FULL) 
				  && (ENTRY_INDEX_IN_BUCKET(d_bucketSize) == 0))  // only bucket group leader should print message
				printf(STATUS_TABLE_FULL);
			elementIndex += stride;
		}
	}
	
	
		// ---


	// -- LITERALS/CONSTANTS --


/** Error message for CUDA errors. */
static const char ERROR_CUDA[] = "Error %s at line %d in file %s\n";


	// -- VARIABLES --


/** Hash table: array of entries. */
inttype *table;


	// -- MACROS --


/** 
 * Helper/inline "function" that checks return values of CUDA calls.
 * Prints any errors and exits application if an error occurred.
 */
#define CUDA_CHECK_RETURN(value)  {                                                        \
	cudaError_t _m_cudaStat = value;                                                       \
	if (_m_cudaStat != cudaSuccess) {                                                      \
		fprintf(stderr, ERROR_CUDA, cudaGetErrorString(_m_cudaStat), __LINE__, __FILE__);  \
		exit(1);                                                                           \
	}                                                                                      \
}

/** Counts all hash table entries for which the associated condition of kernel holds. */
#define COUNT_HOST(kernel)  {                                      \
	indextype result, *count;                                      \
	CUDA_CHECK_RETURN(cudaMallocManaged(&count, sizeof(*count)));  \
	kernel<<<gridDim, blockDim>>>(count);                          \
	CUDA_CHECK_RETURN(cudaDeviceSynchronize());                    \
	result = *count;                                               \
	CUDA_CHECK_RETURN(cudaFree(count));                            \
	return result;                                                 \
}


	// -- FUNCTIONS --


void initialize(indextype h_nEntries, unsigned int h_bucketSize, unsigned int h_vectorLength,
                  unsigned long gridDim, unsigned int blockDim) {
	// allocating table in GPU memory and filling it with empty entries (on GPU):
	CUDA_CHECK_RETURN(cudaMalloc(&table, (h_nEntries * sizeof(*table))));
	CUDA_CHECK_RETURN(cudaMemcpyToSymbol(d_table,       &table,      sizeof(table)));
	CUDA_CHECK_RETURN(cudaMemcpyToSymbol(d_tableLength, &h_nEntries, sizeof(h_nEntries)));
	initTable<<<gridDim, blockDim>>>();
	
	// generating hash constants and copy to GPU constant memory:
	uint64_t h_hashConstants[N_TOTAL_HASH_CONSTANTS];
	init_genrand64(time(NULL));
	for (unsigned int hashFunctionIndex = 0; (hashFunctionIndex < N_TOTAL_HASH_CONSTANTS); hashFunctionIndex++)
		h_hashConstants[hashFunctionIndex] = genrand64_int64();
	CUDA_CHECK_RETURN(cudaMemcpyToSymbol(d_hashConstants, h_hashConstants, sizeof(h_hashConstants)));
	
	// copying (meta) table information to GPU constant memory:
	uint32_t h_nBuckets = (uint32_t) (h_nEntries / h_bucketSize);
	h_bucketSize /= MAX_CHILDREN_BINARY_TREE;  // 32-bit --> 64-bit
	unsigned int h_linearProbeIterations    = (LINEAR_PROBE_LENGTH / h_bucketSize),
	             h_bucketSizeMinLPB         = ((h_bucketSize < LINEAR_PROBE_LENGTH) ? LINEAR_PROBE_LENGTH : h_bucketSize),
	             h_linearExtProbeIterations = (LINEAR_EXT_PROBE_LENGTH / h_bucketSizeMinLPB);
	CUDA_CHECK_RETURN(cudaMemcpyToSymbol(d_nBuckets,                 &h_nBuckets,                 sizeof(h_nBuckets)));
	CUDA_CHECK_RETURN(cudaMemcpyToSymbol(d_bucketSize,               &h_bucketSize,               sizeof(h_bucketSize)));
	CUDA_CHECK_RETURN(cudaMemcpyToSymbol(d_vectorLength,             &h_vectorLength,             sizeof(h_vectorLength)));
	CUDA_CHECK_RETURN(cudaMemcpyToSymbol(d_linearProbeIterations,    &h_linearProbeIterations,    sizeof(h_linearProbeIterations)));
	CUDA_CHECK_RETURN(cudaMemcpyToSymbol(d_bucketSizeMinLPB,         &h_bucketSizeMinLPB,         sizeof(h_bucketSizeMinLPB)));
	CUDA_CHECK_RETURN(cudaMemcpyToSymbol(d_linearExtProbeIterations, &h_linearExtProbeIterations, sizeof(h_linearExtProbeIterations)));
	
	CUDA_CHECK_RETURN(cudaDeviceSynchronize());  // wait till filling table with empty entries is completed
}

void insert(const inttype *h_elements, size_t elementsLength,
              unsigned long gridDim, unsigned int blockDim) {	
	// copying input data to GPU:
	inttype *elements;
	size_t  elementsSize = (elementsLength * sizeof(*h_elements));
	CUDA_CHECK_RETURN(cudaMalloc(&elements, elementsSize));
	CUDA_CHECK_RETURN(cudaMemcpy(elements, h_elements, elementsSize, cudaMemcpyHostToDevice));
	
	// running input data from GPU:
	insertTable<<<gridDim, blockDim>>>(elements, elementsLength);
	CUDA_CHECK_RETURN(cudaDeviceSynchronize());  // wait till insertion is completed
	
	// cleaning up input data:
	CUDA_CHECK_RETURN(cudaFree(elements));
}

indextype elementCount(unsigned long gridDim, unsigned int blockDim) {
	COUNT_HOST(countElements)
}

indextype vectorCount(unsigned long gridDim, unsigned int blockDim) {
	COUNT_HOST(countVectors)
}

void dump(inttype *dest, size_t size) {
	CUDA_CHECK_RETURN(cudaMemcpy(dest, table, size, cudaMemcpyDeviceToHost));
}

void cleanup() {
	CUDA_CHECK_RETURN(cudaFree(table));
	table = NULL;
	CUDA_CHECK_RETURN(cudaDeviceReset());
}