/**
 * GPU hash table implementation.
 * Compressed version (64-bits), with linear probing (32 entries) and fixed bucket size (2).
 *
 * Loosely based on the work of Nathan Cassee (TU/e).
 *
 * Part of Master Project 'Porting tree-based hash table compression to GPGPU model checking', 
 * University of Twente (UT).
 *
 * @author  Danny Bergsma (s0113360)
 * @version 3.0
 */


	// -- INCLUDES --


#include "hash_table.h"

#include <stdio.h>

#include <cuda_runtime.h>

#include "mt64.h"


	// --- DEVICE CODE ---
	
	
		// -- INCLUDES --
	
	
	#include <device_launch_parameters.h>
	
	
		// -- ENUMS --
	
	
	/** Enumeration of possible return values of findOrPut(). */
	enum FoundOrPut {NEW, SEEN, FULL};
	
	
		// -- LITERALS/CONSTANTS --
	
	
	/** 64-bit version of the empty entry. */
	static const uint64_t EMPTY_ENTRY_64    = (((uint64_t) EMPTY_ENTRY_32 << (sizeof(EMPTY_ENTRY_32) * 8)) | EMPTY_ENTRY_32);
	/** Bitmask to unroot 64-bit entries (two 32-bit entries). */
	static const uint64_t UNROOT_BITMASK_64 = ~((uint64_t) ROOT_BIT_32);  // little-endian
	
	/** Number of hash functions (bounded double hashing). */
	static const unsigned int N_HASH_FUNCTIONS              = 32;
	/** Number of constants per hash function. */
	static const unsigned int N_CONSTANTS_PER_HASH_FUNCTION = 3;
	/** Total number of hash constants. */
	static const unsigned int N_TOTAL_HASH_CONSTANTS        = (N_HASH_FUNCTIONS * N_CONSTANTS_PER_HASH_FUNCTION);
	/** Size of hash in bits. */
	static const unsigned int HASH_BITS                     = (sizeof(uint32_t) * 8);

	/** Number of 64-bit entries in a linear probe. */
	static const unsigned int LINEAR_PROBE_LENGTH =     (WARP_SIZE / 8);
	/** Number of 64-bit entries in an external linear probe. */
	static const unsigned int LINEAR_EXT_PROBE_LENGTH = (WARP_SIZE / MAX_CHILDREN_BINARY_TREE);
	
	/** Fixed bucket size. */
	static const unsigned int BUCKET_SIZE = 2;
	
	/** Status message when table is (considered) full. */
	#define STATUS_TABLE_FULL  "Table is (considered) full\n"
	
	
		// -- CONSTANTS --
	
	
	/** Constants used in hash functions. */
	__constant__ uint64_t      d_hashConstants[N_TOTAL_HASH_CONSTANTS];
	/** Hash table: array of hash table entries. */
	__constant__ inttype       *d_table;
	/** 
	 * Number of hash table entries.
	 * @invariant length of d_table array == d_tableLength
	 * @invariant d_tableLength == (d_nBuckets * MAX_CHILDREN_BINARY_TREE)
	 * @invariant (0 < d_tableLength <= MAX_ENTRIES) && ((d_tableLength % WARP_SIZE) == 0)
	 */
	__constant__ indextype     d_tableLength;
	/** Number of buckets. */
	__constant__ uint32_t      d_nBuckets;
	/** Length of vector (number of elements in vectors). */
	__constant__ unsigned int  d_vectorLength;
	
	
		// -- MACROS --
	
	
	/** Counts all hash table entries for which condition holds. */
	#define COUNT_DEVICE(condition)  {                                                                                      \
		indextype localCount = 0;                                                                                           \
		for (indextype entryIndex = ((((indextype) blockIdx.x) * blockDim.x) + threadIdx.x); (entryIndex < d_tableLength);  \
			  (entryIndex += (blockDim.x * gridDim.x)))                                                                     \
			if (condition)                                                                                                  \
				localCount++;                                                                                               \
		                                                                                                                    \
		atomicAdd(count, localCount);                                                                                       \
	}
	
	
		// -- KERNEL: INITIALIZING TABLE --
	
	
	/**
	 * Initializes the hash table by setting all entries to the empty entry value, in parallel.
	 */
	__global__ void initTable() {
		for (indextype entryIndex = ((((indextype) blockIdx.x) * blockDim.x) + threadIdx.x); (entryIndex < d_tableLength);
			  (entryIndex += (blockDim.x * gridDim.x)))  // grid-stride loop
			d_table[entryIndex] = EMPTY_ENTRY_32;
	}
	
	
		// -- KERNEL: COUNTING ELEMENTS --
	
	
	/** 
	 * Counts all elements in hash table, in parallel.
	 * @param count will (eventually) be set to total number of elements in hash table
	 */
	__global__ void countElements(indextype *count) {
		COUNT_DEVICE(d_table[entryIndex] != EMPTY_ENTRY_32)
	}
	
	
		// -- KERNEL: COUNTING VECTORS --
	
	
	/**
	 * Counts all vectors in hash table, in parallel.
	 * @param count will (eventually) be set to total number of vectors in hash table
	 */
	__global__ void countVectors(indextype *count) {
		COUNT_DEVICE(d_table[entryIndex] & ROOT_BIT_32)
	}
	
	
		// -- HELPER FUNCTIONS KERNEL INSERTING VECTORS: FIND_OR_PUT --
	
	
	/**
	 * Returns hash of vector, using the hash function with index hashFunctionIndex.
	 * Hash is valid as index of hash table.
	 * @param    hashFunctionIndex index of hash function to be used
	 * @param    vector            vector to be hashed
	 * @return   hash of vector
	 * @requires hashFunctionIndex < N_HASH_FUNCTIONS
	 * @requires length of vector == MAX_CHILDREN_BINARY_TREE
	 * @ensures  0 <= \result < (d_tableLength / MAX_CHILDREN_BINARY_TREE)
	 */
	__device__ indextype hash(unsigned int hashFunctionIndex, const inttype *vector) {
		unsigned int hashConstantsOffset = (hashFunctionIndex * N_CONSTANTS_PER_HASH_FUNCTION);

		return ((((((d_hashConstants[hashConstantsOffset] + vector[1]) * (d_hashConstants[hashConstantsOffset + 1] + vector[0]))
		           + d_hashConstants[hashConstantsOffset + 2]) >> HASH_BITS) * d_nBuckets) >> HASH_BITS);
	}
	
	/**
	 * Tries to add vector to hash table, using bounded double hashing and linear probing.
	 * Returns index of vector in hash table; EMPTY_ENTRY_32 is returned if table is full.
	 * @param    vector          (sub)vector to be inserted into table
	 * @return   index of vector in hash table; EMPTY_ENTRY_32 is returned if table is full
	 * @requires length of vector == MAX_CHILDREN_BINARY_TREE
	 * @requires (\forall int elementIndex; 0 <= elementIndex < MAX_CHILDREN_BINARY_TREE;
	 *                      (vector[elementIndex] != EMPTY_ENTRY_32))
	 *                         && !(vector[elementIndex] & ROOT_BIT_32)))
	 */
	__device__ indextype findOrPut(const inttype *vector) {
		indextype result = EMPTY_ENTRY_32;  // assume full table (or considered full)
		
		uint64_t vector64 = *((uint64_t*) vector);
		
		unsigned int hashFunctionIndex = 0;
		while ((result == EMPTY_ENTRY_32) && (hashFunctionIndex < N_HASH_FUNCTIONS)) {
			indextype entryIndex = hash(hashFunctionIndex, vector);
			
			unsigned int extProbeIndex = 0;
			do {
				unsigned int probeIndex = 0;
				do {
					uint64_t *entryAddress = (uint64_t*) (d_table + (MAX_CHILDREN_BINARY_TREE * entryIndex)),
							 entry         =             (*entryAddress & UNROOT_BITMASK_64);  // unset root bit
					
					if (entry == vector64)
						result = entryIndex;  // + exit all three loops
					else {  // vector was not found ==> try to insert vector
						entry = (atomicCAS((unsigned long long*) entryAddress, EMPTY_ENTRY_64, vector64)
								 & UNROOT_BITMASK_64);  // unset root bit
						if ((entry == EMPTY_ENTRY_64) || (entry == vector64))
							result = entryIndex;  // + exit all three loops
					}
					
					entryIndex = (((entryIndex / LINEAR_PROBE_LENGTH) * LINEAR_PROBE_LENGTH) + ((entryIndex + 1) % LINEAR_PROBE_LENGTH));
					probeIndex++;
				} while ((result == EMPTY_ENTRY_32) && (probeIndex < LINEAR_PROBE_LENGTH));

				entryIndex = (((entryIndex / LINEAR_EXT_PROBE_LENGTH) * LINEAR_EXT_PROBE_LENGTH)
				              + ((entryIndex + LINEAR_PROBE_LENGTH) % LINEAR_EXT_PROBE_LENGTH));
				extProbeIndex++;
			} while ((result == EMPTY_ENTRY_32) && (extProbeIndex < LINEAR_PROBE_LENGTH));
			
			hashFunctionIndex++;  // try next hash function (double hashing)
		}
		
		return result;
	}
	
	/**
	 * Tries to add (sub)vector to hash table, using tree-based hash table compression. Returns index of vector in hash table.
	 * If table is (considered) full, EMPTY_ENTRY_32 is returned.
	 * @param    vector          (sub)vector to be inserted into table
	 * @param    vectorLength    length of (sub)vector
	 * @return   index of vector in hash table
	 * @requires vectorLength > 0
	 * @requires length of vector == vectorLength
	 * @requires (\forall int elementIndex; 0 <= elementIndex < d_vectorLength;
	 *                      (vector[elementIndex] != EMPTY_ENTRY_32))
	 *                         && !(vector[elementIndex] & ROOT_BIT_32)))
	 */
	__device__ indextype treeRec(const inttype *vector, unsigned int vectorLength) {
		indextype result = EMPTY_ENTRY_32;  // assume full table
		
		indextype recResult[MAX_CHILDREN_BINARY_TREE];
		if (vectorLength < (MAX_CHILDREN_BINARY_TREE * MAX_CHILDREN_BINARY_TREE)) {
		  // vectorLength < 4 ==> we need only one or no recursive call at all
			recResult[0] = vector[0];
			recResult[1] = ((vectorLength == 1) ? 0 
			                                    : ((vectorLength == MAX_CHILDREN_BINARY_TREE) ? vector[1]
			                                                                                  : treeRec((vector + 1), MAX_CHILDREN_BINARY_TREE)));
		} else {  // vectorLength >= 4 ==> two recursive calls
			unsigned int split = (vectorLength / MAX_CHILDREN_BINARY_TREE);
			recResult[0]       = treeRec( vector,          split);
			recResult[1]       = treeRec((vector + split), (vectorLength - split));
		}
		
		if ((recResult[0] != EMPTY_ENTRY_32) && (recResult[1] != EMPTY_ENTRY_32))  // check for full table
			result = findOrPut(recResult);
		
		return result;
	}
	
	/** 
	 * Tries to add vector to hash table, using tree-based hash table compression. 
	 * If it was already present, SEEN is returned. If it was not present, the vector is added and NEW is returned.
	 * If table is (considered) full, FULL is returned.
	 * @param    vector          vector to be inserted into table
	 * @return   result: SEEN (already present), NEW (was not present) or FULL (table is full)
	 * @requires d_vectorLength > 0
	 * @requires length of vector == d_vectorLength 
	 */
	__device__ FoundOrPut treeFindOrPut(const inttype *vector) {
		FoundOrPut result = FULL;  // assume full table
		
		indextype rootIndex = treeRec(vector, d_vectorLength);
		if (rootIndex != EMPTY_ENTRY_32)
			result = ((atomicOr((d_table + (MAX_CHILDREN_BINARY_TREE * rootIndex)), ROOT_BIT_32) & ROOT_BIT_32) ? SEEN : NEW);
			  // check and set ROOT_BIT
		
		return result;
	}
	
	
		// -- KERNEL: INSERTING VECTORS --
	
	
	/**
	 * Inserts the given (elements of) vectors into the (initialized) hash table, in parallel.
	 * Some vectors may not be inserted, if the hash table is (considered) full.
	 * In that case a message will be printed.
	 * @param    d_elements     array of (the elements of) the vectors to be inserted
	 * @param    elementsLength total number of elements to be inserted
	 * @requires (elementsLength % vector length) == 0
	 * @requires elementsLength <= length of d_elements array
	 * @requires (\forall int elementIndex; 0 <= elementIndex < elementsLength; 
	 *                      (d_elements[elementIndex] != EMPTY_ENTRY_32))
	 *                         && !(d_elements[elementIndex] & ROOT_BIT_32)))
	 */
	__global__ void insertTable(const inttype *d_elements, size_t elementsLength) {		
		// processing (elements of) vectors (by grid-stride loop):
		size_t vectorIndex  = ((((size_t) blockIdx.x) * blockDim.x) + threadIdx.x),
		       stride       = ((((size_t) gridDim.x)  * blockDim.x) * d_vectorLength),
		       elementIndex = (vectorIndex * d_vectorLength);
		while (elementIndex < elementsLength) {
			if (treeFindOrPut((d_elements + elementIndex)) == FULL) 
				printf(STATUS_TABLE_FULL);
			elementIndex += stride;
		}
	}
	
	
		// ---


	// -- LITERALS/CONSTANTS --


/** Error message for CUDA errors. */
static const char ERROR_CUDA[] = "Error %s at line %d in file %s\n";


	// -- VARIABLES --


/** Hash table: array of entries. */
inttype *table;


	// -- MACROS --


/** 
 * Helper/inline "function" that checks return values of CUDA calls.
 * Prints any errors and exits application if an error occurred.
 */
#define CUDA_CHECK_RETURN(value)  {                                                        \
	cudaError_t _m_cudaStat = value;                                                       \
	if (_m_cudaStat != cudaSuccess) {                                                      \
		fprintf(stderr, ERROR_CUDA, cudaGetErrorString(_m_cudaStat), __LINE__, __FILE__);  \
		exit(1);                                                                           \
	}                                                                                      \
}

/** Counts all hash table entries for which the associated condition of kernel holds. */
#define COUNT_HOST(kernel)  {                                      \
	indextype result, *count;                                      \
	CUDA_CHECK_RETURN(cudaMallocManaged(&count, sizeof(*count)));  \
	kernel<<<gridDim, blockDim>>>(count);                          \
	CUDA_CHECK_RETURN(cudaDeviceSynchronize());                    \
	result = *count;                                               \
	CUDA_CHECK_RETURN(cudaFree(count));                            \
	return result;                                                 \
}


	// -- FUNCTIONS --


void initialize(indextype h_nEntries, unsigned int h_bucketSize, unsigned int h_vectorLength,
                  unsigned long gridDim, unsigned int blockDim) {
	// allocating table in GPU memory and filling it with empty entries (on GPU):
	CUDA_CHECK_RETURN(cudaMalloc(&table, (h_nEntries * sizeof(*table))));
	CUDA_CHECK_RETURN(cudaMemcpyToSymbol(d_table,       &table,      sizeof(table)));
	CUDA_CHECK_RETURN(cudaMemcpyToSymbol(d_tableLength, &h_nEntries, sizeof(h_nEntries)));
	initTable<<<gridDim, blockDim>>>();
	
	// generating hash constants and copy to GPU constant memory:
	uint64_t h_hashConstants[N_TOTAL_HASH_CONSTANTS];
	init_genrand64(time(NULL));
	for (unsigned int hashFunctionIndex = 0; (hashFunctionIndex < N_TOTAL_HASH_CONSTANTS); hashFunctionIndex++)
		h_hashConstants[hashFunctionIndex] = genrand64_int64();
	CUDA_CHECK_RETURN(cudaMemcpyToSymbol(d_hashConstants, h_hashConstants, sizeof(h_hashConstants)));
	
	// copying (meta) table information to GPU constant memory:
	uint32_t h_nBuckets = (uint32_t) (h_nEntries / BUCKET_SIZE);
	CUDA_CHECK_RETURN(cudaMemcpyToSymbol(d_nBuckets,     &h_nBuckets,     sizeof(h_nBuckets)));
	CUDA_CHECK_RETURN(cudaMemcpyToSymbol(d_vectorLength, &h_vectorLength, sizeof(h_vectorLength)));
	
	CUDA_CHECK_RETURN(cudaDeviceSynchronize());  // wait till filling table with empty entries is completed
}

void insert(const inttype *h_elements, size_t elementsLength,
              unsigned long gridDim, unsigned int blockDim) {	
	// copying input data to GPU:
	inttype *elements;
	size_t  elementsSize = (elementsLength * sizeof(*h_elements));
	CUDA_CHECK_RETURN(cudaMalloc(&elements, elementsSize));
	CUDA_CHECK_RETURN(cudaMemcpy(elements, h_elements, elementsSize, cudaMemcpyHostToDevice));
	
	// running input data from GPU:
	insertTable<<<gridDim, blockDim>>>(elements, elementsLength);
	CUDA_CHECK_RETURN(cudaDeviceSynchronize());  // wait till insertion is completed
	
	// cleaning up input data:
	CUDA_CHECK_RETURN(cudaFree(elements));
}

indextype elementCount(unsigned long gridDim, unsigned int blockDim) {
	COUNT_HOST(countElements)
}

indextype vectorCount(unsigned long gridDim, unsigned int blockDim) {
	COUNT_HOST(countVectors)
}

void dump(inttype *dest, size_t size) {
	CUDA_CHECK_RETURN(cudaMemcpy(dest, table, size, cudaMemcpyDeviceToHost));
}

void cleanup() {
	CUDA_CHECK_RETURN(cudaFree(table));
	table = NULL;
	CUDA_CHECK_RETURN(cudaDeviceReset());
}