/**
 * Header file for hash_table (GPU hash table implementation).
 * Uncompressed version.
 *
 * Loosely based on the work of Nathan Cassee (TU/e).
 *
 * Part of Master Project 'Porting tree-based hash table compression to GPGPU model checking', 
 * University of Twente (UT).
 *
 * @author  Danny Bergsma (s0113360)
 * @version 3.0
 */

#ifndef HASH_TABLE_H
#define HASH_TABLE_H


	// -- INCLUDES--


#include <stdint.h>


	// -- TYPEDEFS --


/** Type of elements in hash table. */
typedef uint32_t inttype;
/** Type of indices in hash table. */
typedef uint64_t indextype;


	// -- LITERALS/CONSTANTS --


/** Number of threads in a warp. */
static const unsigned int WARP_SIZE = 32;

/** Restricted value that indicates an empty entry in hash table. */
static const inttype      EMPTY_ENTRY       = ~((inttype) 0);
/** Size of inttype in bits. */
static const unsigned int INTTYPE_SIZE_BITS = (sizeof(inttype) * 8);
/** Indicates restricted bit for 'writing bucket in progress'. */
static const inttype      WRITING_BIT       = (((inttype) 1) << (INTTYPE_SIZE_BITS - 1));

/** Size of hash in bits. */
static const unsigned int       HASH_BITS   = (sizeof(uint32_t) * 8);
/** Maximum number of buckets. */
static const unsigned long long MAX_BUCKETS = (1ull << HASH_BITS);

/** Maximum grid dimension (number of blocks). */
static const unsigned long MAX_GRID_DIM  = 2147483647ul;
/** Maximum block dimension (number of threads in a block). */
static const unsigned int  MAX_BLOCK_DIM = 1024;


	// -- FUNCTION DECLARATIONS --


/**
 * Initializes the (GPU) hash table.
 * @param    h_nEntries     number of (inttype) entries in hash table
 * @param    h_bucketSize   bucket size (number of hash table entries that form one bucket)
 * @param    h_vectorLength length of vector (number of elements)
 * @param    gridDim        grid dimension (number of blocks)
 * @param    blockDim       block dimension (number of threads in a block)
 * @requires (0 < h_nEntries <= (MAX_BUCKETS * h_bucketSize)) && ((h_nEntries % WARP_SIZE) == 0)
 * @requires (0 < h_bucketSize <= WARP_SIZE) && (h_bucketSize is power of 2)
 * @requires (0 < gridDim <= MAX_GRID_DIM) && (0 < blockDim <= MAX_BLOCK_DIM)
 * @requires h_vectorLength <= h_bucketSize
 */
void initialize(indextype h_nEntries, unsigned int h_bucketSize, unsigned int h_vectorLength,
                  unsigned long gridDim, unsigned int blockDim);

/**
 * Inserts the given (elements of) vectors into the (initialized) hash table, in parallel.
 * Some vectors may not be inserted, if the hash table is (considered) full. 
 * In that case a message will be printed.
 * @param    h_elements     array of the elements of the vectors to be inserted
 * @param    elementsLength total number of elements to be inserted
 * @param    gridDim        grid dimension (number of blocks)
 * @param    blockDim       block dimension (number of threads in a block)
 * @requires (elementsLength % vector length) == 0
 * @requires elementsLength <= length of h_elements array
 * @requires (\forall int elementIndex; (0 <= elementIndex < elementsLength);
 *                      ((h_elements[elementIndex] != EMPTY_ENTRY)
 *                         && !(h_elements[elementIndex] & WRITING_BIT)))
 * @requires (0 < gridDim <= MAX_GRID_DIM) && (0 < blockDim <= MAX_BLOCK_DIM)
 * @requires (blockDim % WARP_SIZE) == 0
 */
void insert(const inttype *h_elements, size_t elementsLength,
              unsigned long gridDim, unsigned int blockDim);

/**
 * Returns number of elements (non-empty entries) in (initialized) hash table, in parallel.
 * @param    gridDim  grid dimension (number of blocks)
 * @param    blockDim block dimension (number of threads in a block)
 * @return   number of elements in hash table
 * @requires (0 < gridDim <= MAX_GRID_DIM) && (blockDim <= MAX_BLOCK_DIM)
 */
indextype elementCount(unsigned long gridDim, unsigned int blockDim);

/**
 * Saves dump of (initialized) hash table to dest.
 * @param    dest destination of dump
 * @param    size size of dump
 * @requires size <= size of dest
 */ 
void dump(inttype *dest, size_t size);

/**
 * Cleans up the (initialized) hash table. Hash table is then not initialized anymore.
 */
void cleanup();

#endif  // !HASH_TABLE_H