/**
 * GPU hash table implementation.
 * Uncompressed version.
 *
 * Stand-alone implementation of hash table implementation from GPUexplore (v2), 
 * but with configurable bucket size.
 * Loosely based on the work of Nathan Cassee (TU/e).
 *
 * Part of Master Project 'Implementing tree-based hash table compression in GPGPU model checking', 
 * University of Twente (UT).
 *
 * @author  Danny Bergsma (s0113360)
 * @version 2.0
 */


	// -- INCLUDES --


#include "hash_table.h"

#include <stdio.h>

#include <cuda_runtime.h>


	// --- DEVICE CODE ---
	
	
		// -- INCLUDES --
	
	
	#include <device_launch_parameters.h>
	
	
		// -- ENUMS --
	
	
	/** Enumeration of possible return values of findOrPut(). */
	enum FoundOrPut {NEW, SEEN, FULL};
	
	
		// -- LITERALS/CONSTANTS --
	
	
	/** Number of hash functions (bounded double hashing). */
	static const unsigned int       N_HASH_FUNCTIONS              = 32;
	/** Number of constants per hash function. */
	static const unsigned int       N_CONSTANTS_PER_HASH_FUNCTION = 2;
	/** Total number of hash constants. */
	static const unsigned int       N_TOTAL_HASH_CONSTANTS        = (N_HASH_FUNCTIONS * N_CONSTANTS_PER_HASH_FUNCTION);
	/** Number of bits shift in hashing function. */
	static const unsigned int       SHIFT                         = 5;	
	static const unsigned long long PRIME_DIVISOR                 = 2147483647ull;  // Mersenne prime 2^31 - 1
	
	/** Status message when table is (considered) full. */
	#define STATUS_TABLE_FULL  "Table is (considered) full\n"
	
	
		// -- CONSTANTS --
	
	
	/** Constants used in hash functions. */
	__constant__ unsigned int d_hashConstants[N_TOTAL_HASH_CONSTANTS];
	/** Hash table: array of hash table entries. */
	__constant__ inttype      *d_table;
	/** 
	 * Number of hash table entries.
	 * @invariant length of d_table array == d_tableLength
	 * @invariant d_tableLength == (d_nBuckets * d_bucketSize)
	 * @invariant (d_tableLength > 0) && ((d_tableLength % WARP_SIZE) == 0)
	 */
	__constant__ indextype    d_tableLength;
	/** 
	 * Number of buckets. 
	 * @invariant 0 < d_nBuckets < PRIME_DIVISOR
	 */
	__constant__ uint32_t     d_nBuckets;
	/**
	 * Number of hash table entries that form a bucket.
	 * @invariant (0 < d_bucketSize <= WARP_SIZE) && (d_bucketSize is power of 2)
	 */
	__constant__ unsigned int d_bucketSize;
	/**
	 * Number of hash table entries that form a bucket, but not crossing half warp boundary.
	 * @invariant (0 < d_simulatedBucketSize <= HALF_WARP_SIZE) && (d_simulatedBucketSize is power of 2)
	 */
	__constant__ unsigned int d_simulatedBucketSize;
	/**
	 * Length of vector (number of elements in vectors).
	 * @invariant d_vectorLength <= d_bucketSize
	 */
	__constant__ unsigned int d_vectorLength;
	
	
		// -- MACROS --
	
	
	/** 
	 * Index of thread in warp.
	 * @invariant 0 <= LANE < WARP_SIZE
	 */
	#define LANE  (threadIdx.x % WARP_SIZE)
	
	/** 
	 * Index of bucket group (in warp) the current thread belongs to.
	 * @invariant 0 <= BUCKET_GROUP_INDEX < WARP_SIZE
	 */
	#define BUCKET_GROUP_INDEX(bucketSize)  (LANE / bucketSize)
	
	/**
	 * Index (offset) of entry (in hash table bucket) that corresponds to the current thread.
	 * @invariant 0 <= ENTRY_INDEX_IN_BUCKET < bucketSize
	 */
	#define ENTRY_INDEX_IN_BUCKET(bucketSize)               (LANE % bucketSize)
	/**
	 * Index (not offset) of (complete/incomplete) vector slot (in hash table bucket) 
	 * that corresponds to the current thread.
	 * @invariant 0 <= SLOT_INDEX_IN_BUCKET < ceil(bucketSize / vectorLength)
	 */
	#define SLOT_INDEX_IN_BUCKET(bucketSize, vectorLength)  (ENTRY_INDEX_IN_BUCKET(bucketSize) / vectorLength)
	
	/**
	 * Index (offset) of element (in vector) that corresponds to the current thread.
	 * @invariant 0 <= ELEMENT_INDEX < vectorLength
	 */
	#define ELEMENT_INDEX(bucketSize, vectorLength)  (ENTRY_INDEX_IN_BUCKET(bucketSize) % vectorLength)
	
	/** Sets bits lower..(lower+amount-1) of bitmask to 1, all other bits to 0 (index LSB = 0). */
	#define SET_BITS(bitmask, lower, amount)  bitmask = (((1ul << (lower + amount)) - 1) ^ ((1ul << lower) - 1));
	
	/** Returns whether predicate holds for any slot in current bucket. */
	#define CHECK_MATCHING_SLOTS(predicate, vectorGroupMask, bucketGroupMask)                                    \
		__any_sync(bucketGroupMask, __all_sync(vectorGroupMask, predicate))
	/**
	 * Returns lane+1 of the thread which corresponding entry in hash table is the start entry of the
	 * first empty complete slot in the bucket (of this bucket group).
	 * Returns 0 if there are no empty complete slots in the bucket (of this bucket group).
	 * @ensures 0 <= \result <= WARP_SIZE
	 */
	#define GET_FIRST_EMPTY_SLOT(vectorLength, vectorGroupMask, bucketGroupMask)                                 \
		__ffs(__ballot_sync(bucketGroupMask,                                                                     \
		                    (__popc(__ballot_sync(vectorGroupMask, (entry == EMPTY_ENTRY))) == vectorLength)));
	
	
		// -- KERNEL: INITIALIZING TABLE --
	
	
	/**
	 * Initializes the hash table by setting all entries to the empty entry value, in parallel.
	 */
	__global__ void initTable() {
		for (indextype entryIndex = ((blockIdx.x * blockDim.x) + threadIdx.x); (entryIndex < d_tableLength);
		       (entryIndex += (blockDim.x * gridDim.x)))  // grid-stride loop
			d_table[entryIndex] = EMPTY_ENTRY;
	}
	
	
		// -- KERNEL: COUNTING ELEMENTS --
	
	
	/** 
	 * Counts all elements in hash table, in parallel.
	 * @param count will (eventually) be set to total number of elements in hash table
	 */
	__global__ void countElements(indextype *count) {
		indextype localCount = 0;  // thread-local count
		for (indextype entryIndex = ((blockIdx.x * blockDim.x) + threadIdx.x); (entryIndex < d_tableLength);
		       (entryIndex += (blockDim.x * gridDim.x)))  // grid-stride loop
			if (d_table[entryIndex] != EMPTY_ENTRY)
				localCount++;
		
		atomicAdd((unsigned long long*) count, localCount);  // add to global count
	}
	
	
		// -- HELPER FUNCTIONS KERNEL INSERTING VECTORS: FIND_OR_PUT --
	
	
	/**
	 * Returns hash of vector, using the hash function with index hashFunctionIndex.
	 * Hash is valid as index of hash table.
	 * @param    hashFunctionIndex index of hash function to be used
	 * @param    vector            vector to be hashed
	 * @return   hash of vector
	 * @requires hashFunctionIndex < N_HASH_FUNCTIONS
	 * @requires length of vector == d_vectorLength
	 * @ensures  0 <= \result <= (d_tableLength - d_bucketSize)
	 */
	__device__ indextype hash(unsigned int hashFunctionIndex, const inttype *vector) {
		indextype hashTmp = d_vectorLength;
		
		for (unsigned int elementIndex = 0; (elementIndex < d_vectorLength); elementIndex++)
			hashTmp = ((hashTmp + vector[elementIndex]) << SHIFT);
		hashTmp = ((hashTmp * d_hashConstants[hashFunctionIndex * N_CONSTANTS_PER_HASH_FUNCTION])
		             + d_hashConstants[(hashFunctionIndex * N_CONSTANTS_PER_HASH_FUNCTION) + 1]);
		hashTmp = (((indextype) ((hashTmp % PRIME_DIVISOR) % d_nBuckets)) * d_bucketSize);
		
		return hashTmp;
	}
	
	/**
	 * Tries to add vector to hash table, using (partial) warp-the-line and bounded double hashing. 
	 * If it was already present, SEEN is returned. If it was not present, the vector is added and NEW is returned.
	 * If table is (considered) full, FULL is returned.
	 * May result in false negatives (i.e., NEW is returned, but concurrent execution added same vector).
	 * @param    vector          vector to be inserted into table
	 * @param    vectorGroupMask bitmask of vector group the current thread belongs to
	 * @param    bucketGroupMask bitmask of bucket group the current thread belongs to
	 * @return   result: SEEN (already present), NEW (was not present) or FULL (table is full)
	 * @requires length of vector == d_vectorLength
	 * @requires (\forall int elementIndex; 0 <= elementIndex < d_vectorLength; 
	 *                      (vector[elementIndex] != EMPTY_ENTRY))
	 */
	__device__ FoundOrPut findOrPut(const inttype *vector, uint32_t vectorGroupMask, uint32_t bucketGroupMask) {
		FoundOrPut result = FULL;  // assume full table (or considered full)
		
		inttype element = vector[ELEMENT_INDEX(d_simulatedBucketSize, d_vectorLength)];
		
		unsigned int hashFunctionIndex = 0;
		while ((result == FULL) && (hashFunctionIndex < N_HASH_FUNCTIONS)) {
			indextype entryIndex    = (hash(hashFunctionIndex, vector) + ENTRY_INDEX_IN_BUCKET(d_bucketSize));
			inttype   *entryAddress = (d_table + entryIndex);
			inttype   entry         = *entryAddress;
			
			if (CHECK_MATCHING_SLOTS((entry == element), vectorGroupMask, bucketGroupMask))
			  // vector group(s) found complete vector in (slots of bucket of) hash table
				result = SEEN;  // + exit loop
			else {  // vector was not found ==> try to insert vector into first empty slot in bucket
			  // setting firstEmptySlot to lane+1 of leader of first vector group with empty slot:
				unsigned int firstEmptySlot = GET_FIRST_EMPTY_SLOT(d_vectorLength, vectorGroupMask, bucketGroupMask);  // 0..32
				
				while ((result == FULL) && firstEmptySlot) {
				  // continue till vector is inserted or there are no empty (complete) slots anymore
					firstEmptySlot--;  // firstEmptySlot is in 1..32 <-> LANE is in 0..31
					
					// exactly one vector group tries to claim the slot:
					if ((firstEmptySlot <= LANE) && (LANE < (firstEmptySlot + d_vectorLength)))
						entry = atomicCAS(entryAddress, EMPTY_ENTRY, element);
					
					if (__shfl_sync(bucketGroupMask, entry, firstEmptySlot) == EMPTY_ENTRY)
					  // vector group successfully claimed the slot
						result = NEW;   // + exit both loops
					else if (CHECK_MATCHING_SLOTS((entry == element), vectorGroupMask, bucketGroupMask))
						result = SEEN;  // + exit both loops
					else
						firstEmptySlot = GET_FIRST_EMPTY_SLOT(d_vectorLength, vectorGroupMask, bucketGroupMask);
						  // try next empty slot in bucket (if there is one)
				}
			}
			
			hashFunctionIndex++;  // try next hash function (double hashing)
		}
		
		return result;
	}
	
	
		// -- KERNEL: INSERTING VECTORS --
	
	
	/**
	 * Inserts the given (elements of) vectors into the (initialized) hash table, in parallel.
	 * Some vectors may not be inserted, if the hash table is (considered) full.
	 * In that case a message will be printed.
	 * Some vectors may appear multiple times in hash table.
	 * @param    d_elements     array of (the elements of) the vectors to be inserted
	 * @param    elementsLength total number of elements to be inserted
	 * @requires (elementsLength % vector length) == 0
	 * @requires elementsLength <= length of d_elements array
	 * @requires (\forall int elementIndex; 0 <= elementIndex < elementsLength; 
	 *                      (d_elements[elementIndex] != EMPTY_ENTRY))
	 */
	__global__ void insertTable(const inttype *d_elements, size_t elementsLength) {
		uint32_t bucketGroupMask, simulatedBucketGroupMask, vectorGroupMask;
		/* setting bitmask to bucket group of current thread: only the bits of those threads are set 
		   that are part of the bucket group of the current thread */
		SET_BITS(bucketGroupMask, (BUCKET_GROUP_INDEX(d_bucketSize) * d_bucketSize), d_bucketSize)
		SET_BITS(simulatedBucketGroupMask, 
		         (BUCKET_GROUP_INDEX(d_simulatedBucketSize) * d_simulatedBucketSize),
		         d_simulatedBucketSize)
		/* setting bitmask to vector group of current thread: only the bits of those threads are set
		   that are part of the vector group of the current thread */
		SET_BITS(vectorGroupMask, 
		         ((BUCKET_GROUP_INDEX(d_simulatedBucketSize) * d_simulatedBucketSize) 
		          + (SLOT_INDEX_IN_BUCKET(d_simulatedBucketSize, d_vectorLength) * d_vectorLength)),
		         d_vectorLength)
		vectorGroupMask &= simulatedBucketGroupMask;  // restrict vector groups to bucket group
		
		// processing (elements of) vectors (by grid-stride loop):
		size_t       vectorIndex  = (((blockIdx.x * blockDim.x) + threadIdx.x) / d_bucketSize);
		unsigned int stride       = (((gridDim.x * blockDim.x) / d_bucketSize) * d_vectorLength);
		size_t       elementIndex = (vectorIndex * d_vectorLength);
		while (elementIndex < elementsLength) {
			if ((findOrPut((d_elements + elementIndex), vectorGroupMask, bucketGroupMask) == FULL) 
				  && (ENTRY_INDEX_IN_BUCKET(d_bucketSize) == 0))  // only bucket group leader should print message
				printf(STATUS_TABLE_FULL);
			elementIndex += stride;
		}
	}
	
	
		// ---


	// -- LITERALS/CONSTANTS --


/** Seed for generating (pseudo)random hash constants. */
static const unsigned int SEED = 4172;

/** Number of threads in a half warp. */
static const unsigned int HALF_WARP_SIZE = (WARP_SIZE / 2);

/** Error message for CUDA errors. */
static const char ERROR_CUDA[] = "Error %s at line %d in file %s\n";


	// -- VARIABLES --


/** Hash table: array of entries. */
inttype *table;


	// -- MACROS --


/** 
 * Helper/inline "function" that checks return values of CUDA calls.
 * Prints any errors and exits application if an error occurred.
 */
#define CUDA_CHECK_RETURN(value)  {                                                        \
	cudaError_t _m_cudaStat = value;                                                       \
	if (_m_cudaStat != cudaSuccess) {                                                      \
		fprintf(stderr, ERROR_CUDA, cudaGetErrorString(_m_cudaStat), __LINE__, __FILE__);  \
		exit(1);                                                                           \
	}                                                                                      \
}


	// -- FUNCTIONS --


void initialize(indextype h_nEntries, unsigned int h_bucketSize, unsigned int h_vectorLength,
                  unsigned long gridDim, unsigned int blockDim) {
	// allocating table in GPU memory and filling it with empty entries (on GPU):
	CUDA_CHECK_RETURN(cudaMalloc(&table, (h_nEntries * sizeof(*table))));
	CUDA_CHECK_RETURN(cudaMemcpyToSymbol(d_table,       &table,      sizeof(table)));
	CUDA_CHECK_RETURN(cudaMemcpyToSymbol(d_tableLength, &h_nEntries, sizeof(h_nEntries)));
	initTable<<<gridDim, blockDim>>>();
	
	// generating hash constants and copy to GPU constant memory:
	unsigned int h_hashConstants[N_TOTAL_HASH_CONSTANTS];
	srand(SEED);
	for (unsigned int hashFunctionIndex = 0; (hashFunctionIndex < N_TOTAL_HASH_CONSTANTS); hashFunctionIndex++)
		h_hashConstants[hashFunctionIndex] = rand();
	CUDA_CHECK_RETURN(cudaMemcpyToSymbol(d_hashConstants, h_hashConstants, sizeof(h_hashConstants)));
	
	// copying (meta) table information to GPU constant memory:
	uint32_t h_nBuckets = (uint32_t) (h_nEntries / h_bucketSize);
	CUDA_CHECK_RETURN(cudaMemcpyToSymbol(d_nBuckets,     &h_nBuckets,     sizeof(h_nBuckets)));
	CUDA_CHECK_RETURN(cudaMemcpyToSymbol(d_bucketSize,   &h_bucketSize,   sizeof(h_bucketSize)));
	CUDA_CHECK_RETURN(cudaMemcpyToSymbol(d_vectorLength, &h_vectorLength, sizeof(h_vectorLength)));
	unsigned int h_simulatedBucketSize = h_bucketSize;
	if (h_bucketSize > HALF_WARP_SIZE)
		h_simulatedBucketSize = HALF_WARP_SIZE;
	CUDA_CHECK_RETURN(cudaMemcpyToSymbol(d_simulatedBucketSize, &h_simulatedBucketSize, sizeof(h_simulatedBucketSize)));
	
	CUDA_CHECK_RETURN(cudaDeviceSynchronize());  // wait till filling table with empty entries is completed
}

void insert(const inttype *h_elements, size_t elementsLength,
              unsigned long gridDim, unsigned int blockDim) {	
	// copying input data to GPU:
	inttype *elements;
	size_t  elementsSize = (elementsLength * sizeof(*h_elements));
	CUDA_CHECK_RETURN(cudaMalloc(&elements, elementsSize));
	CUDA_CHECK_RETURN(cudaMemcpy(elements, h_elements, elementsSize, cudaMemcpyHostToDevice));
	
	// running input data from GPU:
	insertTable<<<gridDim, blockDim>>>(elements, elementsLength);
	CUDA_CHECK_RETURN(cudaDeviceSynchronize());  // wait till insertion is completed
	
	// cleaning up input data:
	CUDA_CHECK_RETURN(cudaFree(elements));
}

indextype elementCount(unsigned long gridDim, unsigned int blockDim) {
	// using Unified Memory/Unified Virtual Addressing:
	indextype result, *count;
	CUDA_CHECK_RETURN(cudaMallocManaged(&count, sizeof(*count)));
	countElements<<<gridDim, blockDim>>>(count);
	CUDA_CHECK_RETURN(cudaDeviceSynchronize());  // wait till counting is completed
	result = *count;
	CUDA_CHECK_RETURN(cudaFree(count));
	return result;
}

void dump(inttype *dest, size_t size) {
	CUDA_CHECK_RETURN(cudaMemcpy(dest, table, size, cudaMemcpyDeviceToHost));
}

void cleanup() {
	CUDA_CHECK_RETURN(cudaFree(table));
	table = NULL;
	CUDA_CHECK_RETURN(cudaDeviceReset());
}